#!/bin/sh
### BEGIN INIT INFO
# Provides:          patd
# Required-Start:    $local_fs $network $remote_fs $syslog
# Required-Stop:     $local_fs $network $remote_fs $syslog
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: Ping and Traceroute Daemon for IPv4 and IPv6
# Description:       "patd" is a daemon to ping and traceroute remote hosts. It is a daemon that
#                    runs without root privileges listening on a TCP port for hosts to check.
#                    "patd" is useful in situations where you cannot send native ICMP packet, for
#                    example from a Java program.
#                    "pat" is a client that talks to patd.
### END INIT INFO

# Author: Maurizio Sartori <masar@Masarlabs.com>

PATH=/sbin:/usr/sbin:/bin:/usr/bin
DESC="patd"
NAME="patd"
PIDDIR="/var/run/$NAME"
PIDFILE="$PIDDIR/$NAME.pid"
DAEMON="/usr/sbin/patd"
DAEMON_ARGS="-p $PIDFILE"
SCRIPTNAME="$0"

[ -x "$DAEMON" ] || exit 0
[ -r "/etc/default/$NAME" ] && . "/etc/default/$NAME"
. /lib/init/vars.sh
. /lib/lsb/init-functions

do_start()
{
    mkdir -p "$PIDDIR" && chown patd.patd "$PIDDIR"
    start-stop-daemon -S -q -p "$PIDFILE" -x "$DAEMON" -c patd:patd -t > /dev/null || return 1
    start-stop-daemon -S -q -p "$PIDFILE" -x "$DAEMON" -c patd:patd -- $DAEMON_ARGS || return 2
}

do_stop()
{
    start-stop-daemon -K -q -R TERM/30/KILL/5 -p "$PIDFILE" -n "$NAME"
    RETVAL="$?"
    [ "$RETVAL" = 2 ] && return 2

    start-stop-daemon -K -q -o -R 0/30/KILL/5 -x "$DAEMON"
    [ "$?" = 2 ] && return 2

    rm -f "$PIDFILE"
    return "$RETVAL"
}

case "$1" in
    start)
        [ "$VERBOSE" != no ] && log_daemon_msg "Starting $DESC" "$NAME"
        do_start
        case "$?" in
            0|1) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
            2)   [ "$VERBOSE" != no ] && log_end_msg 1 ;;
        esac
        ;;
    stop)
        [ "$VERBOSE" != no ] && log_daemon_msg "Stopping $DESC" "$NAME"
        do_stop
        case "$?" in
            0|1) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
            2)   [ "$VERBOSE" != no ] && log_end_msg 1 ;;
        esac
        ;;
    status)
        status_of_proc "$DAEMON" "$NAME" && exit 0 || exit $?
        ;;
    restart|force-reload)
        log_daemon_msg "Restarting $DESC" "$NAME"
        do_stop
        case "$?" in
            0|1)
                do_start
                case "$?" in
                    0) log_end_msg 0 ;;
                    1) log_end_msg 1 ;; # Old process is still running
                    *) log_end_msg 1 ;; # Failed to start
                esac
                ;;
            *)
                # Failed to stop
                log_end_msg 1
                ;;
        esac
        ;;
    *)
        echo "Usage: $SCRIPTNAME {start|stop|status|restart|force-reload}" >&2
        exit 3
        ;;
esac

:
