/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: ip4.c 4108 2014-02-06 17:10:52Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: ip4.c 4108 2014-02-06 17:10:52Z masar@masar $";

#include "ip4.h"
#include "network.h"
#include "util.h"

#include <errno.h>
#include <netinet/ip_icmp.h>
#include <string.h>
#include <syslog.h>

//************************************************************************************************/
//* Data
//************************************************************************************************/
#define IP_HLEN(ip) ((ip).ihl * 4)

static void ip4SendIcmpEcho (PatdClient * pd);

/*************************************************************************************************/
/** Writes IP4 response.
**************************************************************************************************/
static int ip4SendClientResponse (PatdClient * pd, uint8_t iResult,
                                  uint32_t ui32Addres4, uint8_t iCurSeq, uint32_t lMSec)
   {
   int iCount = 0;
   aui8_Buffer[iCount++] = iResult;
   aui8_Buffer[iCount++] = (iCurSeq     >>  0) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >> 24) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >> 16) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >>  8) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >>  0) & 0xFF;
   aui8_Buffer[iCount++] = (ui32Addres4 >>  0) & 0xFF;
   aui8_Buffer[iCount++] = (ui32Addres4 >>  8) & 0xFF;
   aui8_Buffer[iCount++] = (ui32Addres4 >> 16) & 0xFF;
   aui8_Buffer[iCount++] = (ui32Addres4 >> 24) & 0xFF;

   if (iCount != RESPONSE_IP4_SIZE) {
      syslog(LOG_ERR, "ip4SendClientResponse: invalid size");
      deletePatdClient(pd);
      return 1;
      }

   if (writeAll(pd->fdClient, aui8_Buffer, RESPONSE_IP4_SIZE)) {
      syslog(LOG_WARNING, "ip4SendClientResponse: cannot write to client: socket=%i id=%i",
                          pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return 1;
      }
   return 0;
   }

/*************************************************************************************************/
/** Writes ip4 response.
**************************************************************************************************/
static void ip4WriteClientError (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip4WriteClientError: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_ERR, "ip4WriteClientError: invalid status");
      return;
      }

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult = ((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_ERROR :
                        RESULT_ERROR_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;
   uint32_t ui32Addres4 = ((struct sockaddr_in *) &pd->saAddr)->sin_addr.s_addr;

   if (ip4SendClientResponse(pd, iResult, ui32Addres4, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_ERROR) {
      ip4SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip4WriteClientError: ok");
   }

/*************************************************************************************************/
/** Sends an ICMP4 ECHO.
**************************************************************************************************/
static void ip4SendIcmpEcho (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip4SendIcmpEcho: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_DEBUG, "ip4SendIcmpEcho: Invalid status");
      deletePatdClient(pd);
      return;
      }

   if ((pd->iCurRet >= pd->iMaxRet) || (pd->iCurSeq >= pd->iMaxSeq)) {
      syslog(LOG_DEBUG, "ip4SendIcmpEcho: no more");
      pd->iStatus = STATUS_NEW;                          // ???
      rearmClient(pd);
      return;
      }

   pd->iCurSeq += 1;
   pd->iCurTTL = (pd->iCurTTL >= pd->iMaxTTL) ? pd->iCurTTL : pd->iCurTTL + 1;

   ssize_t iSize = pd->iSize + sizeof(struct icmphdr);
   struct icmphdr * pIcmp = (struct icmphdr *) aui8_Buffer;

   memset(aui8_Buffer, 0, iSize);
   aui8_Buffer[iSize-1] = 0xFF;

   pIcmp->type = ICMP_ECHO;
   pIcmp->code = 0;
   pIcmp->checksum = 0;
   pIcmp->un.echo.id = ui32_Pid;
   pIcmp->un.echo.sequence = (pd->iId << MAX_SEQ_BITS) | pd->iCurSeq;

   pIcmp->checksum = icmpChecksum(aui8_Buffer, iSize);

   int iVal = pd->iFragment ? IP_PMTUDISC_DONT : IP_PMTUDISC_DO;
   if (setsockopt(patd_Icmp4.fdIcmp, IPPROTO_IP, IP_MTU_DISCOVER, &iVal, sizeof(iVal)) < 0) {
      syslog(LOG_WARNING, "ip4SendIcmpEcho: set Socket Frag error %i %s",
                          pd->fdClient, strerror(errno));
      deletePatdClient(pd);
      return;
      }

   if (setsockopt(patd_Icmp4.fdIcmp, IPPROTO_IP, IP_TTL, &pd->iCurTTL, sizeof(pd->iCurTTL)) < 0) {
      syslog(LOG_WARNING, "ip4SendIcmpEcho: set Socket TTL error %i %s",
                          pd->fdClient, strerror(errno));
      deletePatdClient(pd);
      return;
      }

   if (sendto(patd_Icmp4.fdIcmp, aui8_Buffer, iSize, 0,
              (struct sockaddr *) &pd->saAddr, sizeof(pd->saAddr)) != iSize) {
      syslog(LOG_DEBUG, "ip4SendIcmpEcho: write IP4 error %i %s", pd->fdClient, strerror(errno));
      pd->iCurSeq = pd->iMaxSeq;       // Force Last
      ip4WriteClientError(pd);
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   syslog(LOG_DEBUG, "ip4SendIcmpEcho: done echo.id=%x echo.seq=%x",
                     pIcmp->un.echo.id, pIcmp->un.echo.sequence);
   }

/*************************************************************************************************/
/** Reads a Client IP4 Config.
**************************************************************************************************/
void ip4ReadConfig (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip4ReadConfig: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_READ_IP4) {
      syslog(LOG_ERR, "ip4ReadConfig: not STATUS_READ_IP4");
      return;
      }

   ssize_t iSize = recv(pd->fdClient, aui8_Buffer, sizeof(aui8_Buffer), MSG_PEEK);
   if (iSize < COMMAND_IP4_SIZE-1) {
      syslog(LOG_DEBUG, "ip4ReadConfig: no enought data size=%zi", iSize);
      return;
      }

   if (readAll(pd->fdClient, aui8_Buffer, COMMAND_IP4_SIZE-1)) {
      syslog(LOG_WARNING, "ip4ReadConfig: error read socket=%i id=%i", pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return;
      }

   int iCount = 0;
   pd->iStatus = STATUS_IP4;

   pd->iMaxSeq = aui8_Buffer[iCount++];
   pd->iMaxSeq = (pd->iMaxSeq < 1) || (pd->iMaxSeq > MAX_SEQ) ?  3 : pd->iMaxSeq;
   pd->iCurSeq = 0;

   pd->iMaxRet = aui8_Buffer[iCount++];
   pd->iMaxRet = (pd->iMaxRet < 1) || (pd->iMaxRet > MAX_SEQ) ?  1 : pd->iMaxRet;
   pd->iCurRet = 0;

   pd->iCurTTL = aui8_Buffer[iCount++];
   pd->iCurTTL = ((pd->iCurTTL < 0) || (pd->iCurTTL > MAX_TTL)) ? MAX_TTL : pd->iCurTTL;

   pd->iMaxTTL = aui8_Buffer[iCount++];
   pd->iMaxTTL = ((pd->iMaxTTL < 1) || (pd->iMaxTTL > MAX_TTL)) ? MAX_TTL : pd->iMaxTTL;

   pd->iCurTTL = (pd->iCurTTL > pd->iMaxTTL) ? pd->iMaxTTL : pd->iCurTTL;

   pd->iSize   = (aui8_Buffer[iCount + 0] <<  8) +
                 (aui8_Buffer[iCount + 1] <<  0);
   iCount += 2;
   pd->iSize   = ((pd->iSize < 0) || (pd->iSize > MAX_ICMP_SIZE)) ? 0 : pd->iSize;

   pd->iFragment = aui8_Buffer[iCount++] ? 1 : 0;

   int iTimeOut = (aui8_Buffer[iCount + 0] <<  8) +
                  (aui8_Buffer[iCount + 1] <<  0);
   iCount += 2;
   if ((iTimeOut <= 0) || (iTimeOut > 2000)) {
      iTimeOut = 1000;
      }

   pd->tsTimeOut.tv_sec  = iTimeOut / 1000;
   pd->tsTimeOut.tv_nsec = (iTimeOut % 1000) * 1000000L;

   struct sockaddr_in * saAddr4 = (struct sockaddr_in *) &pd->saAddr;
   saAddr4->sin_family = AF_INET;
   saAddr4->sin_port = 0;
   saAddr4->sin_addr.s_addr = (aui8_Buffer[iCount + 0] <<  0) +
                              (aui8_Buffer[iCount + 1] <<  8) +
                              (aui8_Buffer[iCount + 2] << 16) +
                              (aui8_Buffer[iCount + 3] << 24);
//   iCount += 6;

   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   syslog(LOG_DEBUG, "ip4ReadConfig: socket=%i id=%i status=%i"
                     " ip=%i.%i.%i.%i seq=%i ret=%i curttl=%i maxttl=%i"
                     " size=%i frag=%i to=%li.%09li",
                     pd->fdClient, pd->iId, pd->iStatus,
                     (saAddr4->sin_addr.s_addr >>  0) & 0xFF,
                     (saAddr4->sin_addr.s_addr >>  8) & 0xFF,
                     (saAddr4->sin_addr.s_addr >> 16) & 0xFF,
                     (saAddr4->sin_addr.s_addr >> 24) & 0xFF,
                     pd->iMaxSeq, pd->iMaxRet,
                     pd->iCurTTL, pd->iMaxTTL, pd->iSize, pd->iFragment,
                     pd->tsTimeOut.tv_sec, pd->tsTimeOut.tv_nsec);

   ip4SendIcmpEcho(pd);
   }

/*************************************************************************************************/
/** Writes ip4 response.
**************************************************************************************************/
static void ip4WriteClientOk (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip4WriteClientOk: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_DEBUG, "ip4WriteClientOk: invalid status");
      return;
      }

   if (pd->iCurRet >= pd->iMaxRet) {
      syslog(LOG_DEBUG, "ip4WriteClientOk: yet sent MaxRet");
      return;
      }
   pd->iCurRet += 1;

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult =((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_OK :
                        RESULT_OK_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;
   uint32_t ui32Addres4 = ((struct sockaddr_in *) &pd->saAddr)->sin_addr.s_addr;

   if (ip4SendClientResponse(pd, iResult, ui32Addres4, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_OK) {
      ip4SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip4WriteClientOk: ok");
   }

/*************************************************************************************************/
/** Writes ip4 response.
**************************************************************************************************/
static void ip4WriteClientTrace (PatdClient * pd, uint32_t ui32Addres4)
   {
   syslog(LOG_DEBUG, "ip4WriteClientTrace: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_ERR, "ip4WriteClientTrace: invalid status");
      return;
      }

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult = ((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_TRACE :
                        RESULT_TRACE_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;

   if (ip4SendClientResponse(pd, iResult, ui32Addres4, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_TRACE) {
      ip4SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip4WriteClientTrace: ok");
   }

/*************************************************************************************************/
/** ip4TimeExceeded.
**************************************************************************************************/
static void ip4TimeExceeded (struct iphdr * pIp, struct icmphdr * pIcmpOrg)
   {
   syslog(LOG_DEBUG, "ip4TimeExceeded: start");

   int iId = pIcmpOrg->un.echo.sequence >> MAX_SEQ_BITS;
   int iCurSeq = pIcmpOrg->un.echo.sequence & MAX_SEQ_MASK;

   if (iId >= MAX_CLIENTS) {
      syslog(LOG_WARNING, "ip4TimeExceeded: invalid id: %i", iId);
      return;
      }
   syslog(LOG_DEBUG, "ip4TimeExceeded: id=%i", iId);

   PatdClient * pd = getPatdClient(iId);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_WARNING, "ip4TimeExceeded: strange status socket=%i id=%i status=%i",
                          pd->fdClient, pd->iId, pd->iStatus);
      deletePatdClient(pd);
      return;
      }

   if (pd->iCurSeq != iCurSeq) {
      syslog(LOG_DEBUG, "ip4TimeExceeded: old seq: socket=%i id=%i seq %i!=%i",
                        pd->fdClient, pd->iId, pd->iCurSeq, iCurSeq);
      return;
      }
   ip4WriteClientTrace(pd, pIp->saddr);

   syslog(LOG_DEBUG, "ip4TimeExceeded: ok");
   }

/*************************************************************************************************/
/** ip4EchoReply.
**************************************************************************************************/
static void ip4EchoReply (struct icmphdr * pIcmp)
   {
   syslog(LOG_DEBUG, "ip4EchoReply: start");
//   syslogDump(LOG_DEBUG, "ECHO4 REPLY DUMP: ", pIcmp, iSize);
//   syslog(LOG_DEBUG, "ECHO4 REPLY: %x %x", pIcmp->un.echo.id, pIcmp->un.echo.sequence);

   int iId = pIcmp->un.echo.sequence >> MAX_SEQ_BITS;
   int iCurSeq = pIcmp->un.echo.sequence & MAX_SEQ_MASK;

   if (iId >= MAX_CLIENTS) {
      syslog(LOG_WARNING, "ip4EchoReply: invalid id: %i", iId);
      return;
      }
   syslog(LOG_DEBUG, "ip4EchoReply: id=%i", iId);

   PatdClient * pd = getPatdClient(iId);
   if (pd->iStatus != STATUS_IP4) {
      syslog(LOG_DEBUG, "ip4EchoReply: strange status socket=%i id=%i status=%i",
                        pd->fdClient, pd->iId, pd->iStatus);
      deletePatdClient(pd);
      return;
      }

   if (pd->iCurSeq != iCurSeq) {
      syslog(LOG_DEBUG, "ip4EchoReply: old seq: socket=%i id=%i seq %i!=%i",
                        pd->fdClient, pd->iId, pd->iCurSeq, iCurSeq);
      return;
      }
   ip4WriteClientOk(pd);

   syslog(LOG_DEBUG, "ip4EchoReply: ok");
   }

/*************************************************************************************************/
/** ip4Process.
**************************************************************************************************/
void ip4Process (void)
   {
//   syslog(LOG_DEBUG, "ip4Process: start");
   ssize_t iSize = recv(patd_Icmp4.fdIcmp, aui8_Buffer, sizeof(aui8_Buffer), 0);
   while (iSize > 0) {
      struct iphdr * pIp = (struct iphdr *) aui8_Buffer;
      struct icmphdr * pIcmp = (struct icmphdr *)((uint8_t *) pIp + IP_HLEN(*pIp));
      struct iphdr * pIpOrg = (struct iphdr *)(pIcmp + 1);
      struct icmphdr * pIcmpOrg = (struct icmphdr *)((uint8_t *) pIpOrg + IP_HLEN(*pIpOrg));

      if ((pIcmp->type == ICMP_TIME_EXCEEDED) && (pIcmpOrg->un.echo.id == ui32_Pid)) {
         syslog(LOG_DEBUG, "ip4Process: icmp ICMP_TIME_EXCEEDED");
         ip4TimeExceeded(pIp, pIcmpOrg);
         }
      else if ((pIcmp->type == ICMP_UNREACH) && (pIcmpOrg->un.echo.id == ui32_Pid)) {
         syslog(LOG_DEBUG, "ip4Process: icmp ICMP_UNREACH");
         ip4TimeExceeded(pIp, pIcmpOrg);
         }
      else if ((pIcmp->type == ICMP_ECHOREPLY) && (pIcmp->un.echo.id == ui32_Pid)) {
         syslog(LOG_DEBUG, "ip4Process: icmp ICMP_ECHOREPLY");
         ip4EchoReply(pIcmp);
         }
      else {
         syslog(LOG_DEBUG, "ip4Process: ignore type=%i code=%i id=%i size=%zi",
                           pIcmp->type, pIcmp->code, pIcmp->un.echo.id, iSize);
         }

      iSize = recv(patd_Icmp4.fdIcmp, aui8_Buffer, sizeof(aui8_Buffer), 0);
      }
//   syslog(LOG_DEBUG, "ip4Process: done");
   }

/*************************************************************************************************/
/** Processed an ip4 TimeOut.
**************************************************************************************************/
void ip4TimeOut (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip4TimeOut: socket=%i id=%i status=%i", pd->fdClient, pd->iId, pd->iStatus);
   ip4WriteClientError(pd);
   }

/*************************************************************************************************/
/*************************************************************************************************/
